@@ -84,7 +84,7 @@ namespace sat {
     };
 
     struct no_drat_params : public params_ref {
-        no_drat_params() { set_sym("drat.file", symbol()); }        
+        no_drat_params() { set_bool("drat.disable", true); }
     };
     
     class solver : public solver_core {
@@ -124,7 +124,7 @@ namespace sat {
         clause_vector           m_clauses;
         clause_vector           m_learned;
         unsigned                m_num_frozen;
-        unsigned_vector         m_active_vars, m_free_vars, m_vars_to_reinit;
+        unsigned_vector         m_active_vars, m_free_vars, m_vars_to_free, m_vars_to_reinit;
         vector<watch_list>      m_watches;
         svector<lbool>          m_assignment;
         svector<justification>  m_justification; 
@@ -266,6 +266,11 @@ namespace sat {
         //
         // -----------------------
         void add_clause(unsigned num_lits, literal * lits, sat::status st) override { mk_clause(num_lits, lits, st); }
+        void add_clause(literal l1, literal l2, status st) {
+            literal lits[2] = { l1, l2 };
+            add_clause(2, lits, st);
+        }
+        void add_clause(literal lit, status st) { literal lits[1] = { lit }; add_clause(1, lits, st); }
         bool_var add_var(bool ext) override { return mk_var(ext, true); }
 
         bool_var mk_var(bool ext = false, bool dvar = true);
@@ -283,7 +288,7 @@ namespace sat {
         inline clause_allocator& cls_allocator() { return m_cls_allocator[m_cls_allocator_idx]; }
         inline clause_allocator const& cls_allocator() const { return m_cls_allocator[m_cls_allocator_idx]; }
         inline clause * alloc_clause(unsigned num_lits, literal const * lits, bool learned) { return cls_allocator().mk_clause(num_lits, lits, learned); }
-        inline void     dealloc_clause(clause* c) { cls_allocator().del_clause(c); }
+        inline void dealloc_clause(clause* c) { cls_allocator().del_clause(c); }
         struct cmp_activity;
         void defrag_clauses();
         bool should_defrag();
@@ -369,7 +374,7 @@ namespace sat {
         void set_eliminated(bool_var v, bool f) override;
         bool was_eliminated(literal l) const { return was_eliminated(l.var()); }
         void set_phase(literal l) override { if (l.var() < num_vars()) m_best_phase[l.var()] = m_phase[l.var()] = !l.sign(); }
-        bool_var get_phase(bool_var b) { return m_phase.get(b, false); }
+        bool get_phase(bool_var b) { return m_phase.get(b, false); }
         void move_to_front(bool_var b);
         unsigned scope_lvl() const { return m_scope_lvl; }
         unsigned search_lvl() const { return m_search_lvl; }
@@ -444,9 +449,10 @@ namespace sat {
         void       flush_roots();
         typedef std::pair<literal, literal> bin_clause;
         struct bin_clause_hash { unsigned operator()(bin_clause const& b) const { return b.first.hash() + 2*b.second.hash(); } };
-    protected:
-        watch_list & get_wlist(literal l) { return m_watches[l.index()]; }
-        watch_list const & get_wlist(literal l) const { return m_watches[l.index()]; }
+
+        watch_list const& get_wlist(literal l) const { return m_watches[l.index()]; }
+        watch_list& get_wlist(literal l) { return m_watches[l.index()]; }
+    protected:            
         watch_list & get_wlist(unsigned l_idx) { return m_watches[l_idx]; }
         bool is_marked(bool_var v) const { return m_mark[v]; }
         void mark(bool_var v) { SASSERT(!is_marked(v)); m_mark[v] = true; }
@@ -521,20 +527,22 @@ namespace sat {
 
         unsigned m_conflicts_since_init { 0 };
         unsigned m_restarts { 0 };
-        unsigned m_restart_next_out { 0 };
-        unsigned m_conflicts_since_restart { 0 };
-        bool     m_force_conflict_analysis { false };
-        unsigned m_simplifications { 0 };
-        unsigned m_restart_threshold { 0 };
-        unsigned m_luby_idx { 0 };
-        unsigned m_conflicts_since_gc { 0 };
-        unsigned m_gc_threshold { 0 };
-        unsigned m_defrag_threshold { 0 };
-        unsigned m_num_checkpoints { 0 };
-        double   m_min_d_tk { 0 } ;
-        unsigned m_next_simplify { 0 };
-        bool     m_simplify_enabled { true };
-        bool     m_restart_enabled { true };
+        unsigned m_restart_next_out = 0;
+        unsigned m_conflicts_since_restart = 0;
+        bool     m_force_conflict_analysis = false;
+        unsigned m_simplifications = 0;
+        unsigned m_restart_threshold = 0;
+        unsigned m_luby_idx = 0;
+        unsigned m_conflicts_since_gc = 0;
+        unsigned m_gc_threshold = 0;
+        unsigned m_defrag_threshold = 0;
+        unsigned m_num_checkpoints = 0;
+        double   m_min_d_tk = 0.0 ;
+        unsigned m_next_simplify = 0;
+        double   m_simplify_mult = 1.5;
+        bool     m_simplify_enabled = true;
+        bool     m_restart_enabled = true;
+        bool guess(bool_var next);
         bool decide();
         bool_var next_var();
         lbool bounded_search();
@@ -706,7 +714,6 @@ namespace sat {
         //
         // -----------------------
     public:
-        void set_should_simplify() { m_next_simplify = m_conflicts_since_init; }
         bool_var_vector const& get_vars_to_reinit() const { return m_vars_to_reinit;  }
         bool is_probing() const { return m_is_probing; }
         bool is_free(bool_var v) const { return m_free_vars.contains(v); }
@@ -749,7 +756,7 @@ namespace sat {
 
         u_map<index_set>       m_antecedents;
         literal_vector         m_todo_antecedents;
-        vector<literal_vector> m_binary_clause_graph;
+        // vector<literal_vector> m_binary_clause_graph;
 
         bool extract_assumptions(literal lit, index_set& s);
         
