@@ -45,7 +45,7 @@ public:
         m_arith(m)
     {}
 
-    ~reduce_invertible_tactic() override { }
+    char const* name() const override { return "reduce_invertible"; }
 
     tactic * translate(ast_manager & m) override {
         return alloc(reduce_invertible_tactic, m);
